/* The spec is 7bit ascii */
%option 7bit
/* For speed */
%option batch
/* XML is case insensitive except where explicitly stated in the spec */
%option caseful
/* Since this is in a shared library, the size isn't a worry really */
%option fast

%option noyywrap
/*%option debug*/


/*
   The XML spec defines character ranges in USC-2, but our scanner needs to match these values in UTF-8.
   This program will read the specification and convert the USC-2 range patterns into UTF-8 patterns.

   See xml_scanner.lm for the definition of UTF-8 copied from RFC-2279.
*/
%{

#include <stdio.h>
#include <libc.h>
#include <assert.h>

void mapRange(char *range, unsigned int length);
void mapSingleCharacter(char *character, unsigned int length);
%}

USC2_CHAR  "#x"[0-9a-fA-f]{1,8}
USC2_RANGE "["{USC2_CHAR}-{USC2_CHAR}"]"

%%

{USC2_RANGE} mapRange(yytext, yyleng);
{USC2_CHAR}  mapSingleCharacter(yytext, yyleng);

%%

/*
  The only tricky part about this is that if we have a range that takes multiple
  bytes in UTF-8 and both end points aren't the same byte encoding except
  for the last byte, then care must be taken to avoid specifying the incorrect
  range.

  For example, say that we have the range string "[#x80-#x700]".

    0x80  --> 110[00010] 10[000000] --> C2 80
    0x712 --> 110[11100] 10[010010] --> DC 92

  We cannot write anything as simple as the following pattern:

     [\xC2-\xDC][\x80-\x92]

  Rather, we must split the pattern up, first specifying the possible values
  beginning with the low end byte of 0xC2 (note that since all 'follow' bytes
  must be of the form 10xxxxxx, the max value in those bytes is 0xBF, likewise
  the minimum value is 0x80):

     \xC2[0x80-\xBF]

  The next possible set of subranges are the complete ranges starting from where
  we left off up to the last full subrange before the endpoint:

     [\xC2-\xDB][\x80-\xBF]

  Finally, we specify the last partial range:

     \xDC[\x80-\x92]

  So our final pattern for this example would be:

     (\xC2[0x80-\xBF])|(\xDC[\x80-\x92])|(\xDC[\x80-\x92])


  As a longer example, say we have "[#x8ab-#x4ccb]":

    0x8ab --> 1110[0000] 10[100010] 10[101011] --> E0 A2 AB
   0x4ccb --> 1110[0100] 10[110011] 10[001011] --> E4 B3 8B

  Here we have to finish off each of the follow bytes on the start, getting the patterns:

    \xE0\xA2[\xAB-\xBF]
    \xEO[\xA3-\xBF][\x80-\xBF]

  Then, as above we have the middle complete ranges:

    [\xE0-\xE3][\x80-\xBF][\x80-\xBF]

  Finally, we have to extend to each of the follow bytes on the end point:

    \xE4[\x80-\xB2][\x80-\xBF]
    \xE4\xB3[\x80-\x8B]


*/

unsigned int alreadyEmittedAPattern;  // Used to determine whether we need to emit another '|'

//
// We define a generalization of the encoding for UTF8 where every character is encoded
// in 6 bytes.  The length field counts how many bytes (starting from the END) are used.
// Bytes before that should be ignored and should be zero.
//

#define UTF8_MAX_LEN (6)

typedef struct _utf8 {
  unsigned char length;
  unsigned char bytes[UTF8_MAX_LEN];
} utf8;

utf8 startValues[UTF8_MAX_LEN], endValues[UTF8_MAX_LEN];

void mapIntToUTF8(unsigned int input, utf8 *result)
{
    unsigned char *bytes = &result->bytes[0];

    bytes[0] = 0;
    bytes[1] = 0;
    bytes[2] = 0;
    bytes[3] = 0;
    bytes[4] = 0;
    bytes[5] = 0;

    if (input <= 0x7f) {
        bytes[5] = input;
        result->length = 1;
    } else if (input <= 0x7FF) {
        bytes[5] = ((input >>  0) & 0x3F) | 0x80;
        bytes[4] = (input >> 6) | 0xC0;
        result->length = 2;
    } else if (input <= 0xFFFF) {
        bytes[5] = ((input >>  0) & 0x3F) | 0x80;
        bytes[4] = ((input >>  6) & 0x3F) | 0x80;
        bytes[3] = (input >> 12) | 0xE0;
        result->length = 3;
    } else if (input <= 0x1FFFFF) {
        bytes[5] = ((input >>  0) & 0x3F) | 0x80;
        bytes[4] = ((input >>  6) & 0x3F) | 0x80;
        bytes[3] = ((input >> 12) & 0x3F) | 0x80;
        bytes[2] = (input >> 18) | 0xF0;
        result->length = 4;
    } else if (input <= 0x3FFFFFF) {
        bytes[5] = ((input >>  0) & 0x3F) | 0x80;
        bytes[4] = ((input >>  6) & 0x3F) | 0x80;
        bytes[3] = ((input >> 12) & 0x3F) | 0x80;
        bytes[2] = ((input >> 18) & 0x3F) | 0x80;
        bytes[1] = (input >> 24) | 0xF8;
        result->length = 5;
    } else if (input <= 0x7FFFFFFF) {
        bytes[5] = ((input >>  0) & 0x3F) | 0x80;
        bytes[4] = ((input >>  6) & 0x3F) | 0x80;
        bytes[3] = ((input >> 12) & 0x3F) | 0x80;
        bytes[2] = ((input >> 18) & 0x3F) | 0x80;
        bytes[1] = ((input >> 24) & 0x3F) | 0x80;
        bytes[0] = (input >> 30) | 0xFC;
        result->length = 6;
    } else {
        fprintf(stderr, "mapIntToUTF8: Illegal input value 0x%08x.\n", input);
        exit(1);
    }
}


// Handles the base trival case where we can assume that we can do full ranges
// on each component (and that the two end points are the same length).
void mapSimpleRange(utf8 start, utf8 end)
{
    unsigned int i;

    assert(start.length == end.length);
    assert(start.length <= UTF8_MAX_LEN);

#ifdef DEBUG
    fprintf(stderr, "\t\t");
#endif
    if (alreadyEmittedAPattern) {
        printf("|");
    } else
        alreadyEmittedAPattern = 1;

    printf("(");
    for (i = UTF8_MAX_LEN - start.length; i < UTF8_MAX_LEN; i++) {
        if (start.bytes[i] == end.bytes[i]) {
#ifdef DEBUG
            fprintf(stderr, "\\x%02X", start.bytes[i]);
#endif
            printf("\\x%02X", start.bytes[i]);
        } else {
#ifdef DEBUG
            fprintf(stderr, "[\\x%02X-\\x%02X]", start.bytes[i], end.bytes[i]);
#endif
            printf("[\\x%02X-\\x%02X]", start.bytes[i], end.bytes[i]);
        }
    }
    printf(")");

#ifdef DEBUG
    fprintf(stderr, "\n");
#endif
   
}


// Handles the case in which start->length == end->length.
void mapEqualLengthRange(utf8 start, utf8 end)
{
    unsigned int firstDiffer, i;
    utf8         xStart, xEnd;

    assert(start.length == end.length);
    assert(start.length <= UTF8_MAX_LEN);

#ifdef DEBUG
    fprintf(stderr, "\n\tstart = [%d: %02X %02X %02X %02X %02X %02X]\n\tend   = [%d: %02X %02X %02X %02X %02X %02X]\n",
            start.length, start.bytes[0], start.bytes[1], start.bytes[2], start.bytes[3], start.bytes[4], start.bytes[5],
            end.length, end.bytes[0], end.bytes[1], end.bytes[2], end.bytes[3], end.bytes[4], end.bytes[5]);
#endif

    // Compute how much of the prefix is equal.  For example, if we have
    // [ 0 0 a1 b1 c1 d1 ] and [ 0 0 a1 b1 c2 d2 ], then we will see that the index of the
    // differing byte is 4.
    for (firstDiffer = 0; firstDiffer < UTF8_MAX_LEN; firstDiffer ++) {
        if (start.bytes[firstDiffer] != end.bytes[firstDiffer])
            break;
    }
#ifdef DEBUG
    fprintf(stderr, "\t\tfirstDiffer = %d\n", firstDiffer);
#endif
    if (firstDiffer == UTF8_MAX_LEN || firstDiffer == UTF8_MAX_LEN-1) {
        // The end points are equal, or only different in the least byte.  In either case,
        // the code below isn't needed and doesn't work right for the 5 case anyway.
#ifdef DEBUG
        fprintf(stderr, "\tSimple block:\n");
#endif
        mapSimpleRange(start, end);
        return;
    }


    // Finish off each one of the byte ranges for the non-equal bytes.  For the above example
    // we'd emit the ranges:
    //
    //    [ 0 0 a1 b1 c1 d1 ] .. [ 0 0 a1 b1 c1 BF ]
    //
    // We can't finish off the first non-equal byte since that would make the start value
    // greater than the end value.  Continuing the above example, that means we don't
    // emit the range [ 0 0 a1 b1 c1 80 ] .. [ 0 0 a1 b1 BF BF ] (since c2 is probably
    // less than BF and even if it isn't we'll handle it later, also, for the first
    // byte in any length group, the max is not 0xBF anyway).
    //
    // IMPORTANT: We also avoid invalid UTF-8 characters here by insetting the range on
    // each side by one.  For example, note that the four byte UTF-8 mapping range looks
    // like the following in binary:
    //
    //   0001 0000 --  11110[xxx] 10[x10000] 10[000000] 10[000000] -- F0 90 80 80
    //   001F FFFF --  11110[111] 10[111111] 10[111111] 10[111111] -- F7 BF BF BF
    //
    // That is, if the prefix '11110' is encountered, the next five bits
    // must be at least '000001'.  If they are all zero, the bit stream is
    // invalid!
    //
    // So if we get an input start/end that looks like [ 0 0 a1 b1 c1 d1 ] ..
    // [ 0 0 a2 b2 c2 d2 ], then we will emit the patterns:
    //
    //    a1 b1        c1        [d1-BF]
    //    a1 b1        [c1+1-BF] [80-BF]
    //    a1 [b1+1-BF] [80-BF]   [80-BF]
    //
    //  If we didn't do this and instead did:
    //
    //    a1      b1      c1 [d1-BF]
    //    a1      b1 [c1-BF] [80-BF]
    //    a1 [b1-BF] [80-BF] [80-BF]  **
    //
    //  Then the '**' pattern would end up matching [ a1 b1 80 80 ] even though c1 might be > 80.
    //

// Macros to define the maximum and minimum values for a particular position
#define MIN_AT_POS(pos)  (startValues[start.length-1].bytes[pos])
#define MAX_AT_POS(pos)  (endValues[end.length-1].bytes[pos])

#define FOLLOW_BYTE_MIN (0x80)
#define FOLLOW_BYTE_MAX (0xBF)

#ifdef DEBUG
    fprintf(stderr, "\tStart blocks:\n");
#endif
    xEnd = xStart = start;
    for (i = 5; i > firstDiffer; i--) {
#ifdef DEBUG
        fprintf(stderr, "\t\ti = %d\n", i);
#endif
        xEnd.bytes[i] = MAX_AT_POS(i);
        mapSimpleRange(xStart, xEnd);
        xStart.bytes[i] = MIN_AT_POS(i);
        if (xStart.bytes[i-1] < MAX_AT_POS(i-1))
            xStart.bytes[i-1]++;
    }

    // Now, map the full internal ranges.  For the above example, this would be:
    //
    //     a1 b1 [(c1+1)-(c2-1)] [0x80-0xBF]
    //
#ifdef DEBUG
    fprintf(stderr, "\tInternal blocks:\n");
#endif
    if (end.bytes[firstDiffer] - start.bytes[firstDiffer] > 1) {
        xStart = start;
        xEnd = end;

        xStart.bytes[firstDiffer]++;
        xEnd.bytes[firstDiffer]--;
        for (i = firstDiffer + 1; i < UTF8_MAX_LEN; i++) {
            xStart.bytes[i] = FOLLOW_BYTE_MIN;
            xEnd.bytes[i]   = FOLLOW_BYTE_MAX;
        }
        mapSimpleRange(xStart, xEnd);
    }

    // Finally, start the end ranges.  For the above example, this would be:
    //
    //   a1 b1 c2 [0x80-d2]
    //
    // Like the above, care must be taken to avoid permitting invalid UTF-8 streams
    // to pass undetected.
#ifdef DEBUG
    fprintf(stderr, "\tEnd blocks:\n");
#endif
    xStart = end;
    xEnd = end;
    for (i = 5; i > firstDiffer; i--) {
#ifdef DEBUG
        fprintf(stderr, "\t\ti = %d\n", i);
#endif
        xStart.bytes[i] = FOLLOW_BYTE_MIN;
        mapSimpleRange(xStart, xEnd);
        xEnd.bytes[i] = FOLLOW_BYTE_MAX;
        if (xEnd.bytes[i-1] > FOLLOW_BYTE_MIN)
            xEnd.bytes[i-1]--;
    }
}


void mapRange(char *range, unsigned int length)
{
    unsigned int startInt, endInt;
    utf8         start, end, x;

    // Possibly a bit nasty
    range[length] = '\0';

    if (sscanf(range, "[#x%x-#x%x]", &startInt, &endInt) != 2) {
        fprintf(stderr, "Unable to extract start/end values from range '%s'.\n", range);
        exit(1);
    }

    if (startInt > endInt) {
        fprintf(stderr, "Bogus range: '%s'.  start must be less than or equal to end.\n", range);
        exit(1);
    }

#ifdef DEBUG
    fprintf(stderr, "range %s: start = %d end = %d\n", range, startInt, endInt);
#endif

    mapIntToUTF8(startInt, &start);
    mapIntToUTF8(endInt, &end);

#ifdef DEBUG
    fprintf(stderr, "\tstart = [%d: %02X %02X %02X %02X %02X %02X]\n\tend   = [%d: %02X %02X %02X %02X %02X %02X]\n",
            start.length, start.bytes[0], start.bytes[1], start.bytes[2], start.bytes[3], start.bytes[4], start.bytes[5],
            end.length, end.bytes[0], end.bytes[1], end.bytes[2], end.bytes[3], end.bytes[4], end.bytes[5]);
#endif

    
    alreadyEmittedAPattern = 0;
    printf("(");

    // Finish off the full ranges of each length.  For example, if we have
    // a starting utf8 value of [C2 80] and and ending one of [E0 A2 AB],
    // then we will emit ranges for [C2 80] to [DF BF] (the max 2 byte value)
    // and then proceed to handle range from the minimum three byte value
    // to the give three by end value.
    x = start;
    while (x.length < end.length) {
        mapEqualLengthRange(x, endValues[x.length - 1]);
        x = startValues[x.length];
    }

    // Map the leftover part (where the endpoints must be of equal lengths)
    assert(x.length == end.length);
    mapEqualLengthRange(x, end);

    printf(")");
}

void mapSingleCharacter(char *character, unsigned int length)
{
    unsigned int intValue;
    utf8         value;

    // Possibly a bit nasty
    character[length] = '\0';

    if (sscanf(character, "#x%x", &intValue) != 1) {
        fprintf(stderr, "Unable to extract integer value character '%s'.\n", character);
        exit(1);
    }

#ifdef DEBUG
    fprintf(stderr, "character %s: int value = %d\n", character, intValue);
#endif

    mapIntToUTF8(intValue, &value);

    alreadyEmittedAPattern = 0;
    printf("(");
    mapSimpleRange(value, value);
    printf(")");
}


void main()
{
#ifdef DEBUG
    int i;
#endif

    // Set up the shared constants
    mapIntToUTF8(0x0,  &startValues[0]);
    mapIntToUTF8(0x7F, &endValues[0]);

    mapIntToUTF8(0x80,  &startValues[1]);
    mapIntToUTF8(0x7FF, &endValues[1]);

    mapIntToUTF8(0x800,  &startValues[2]);
    mapIntToUTF8(0xFFFF, &endValues[2]);

    mapIntToUTF8(0x10000,  &startValues[3]);
    mapIntToUTF8(0x1FFFFF, &endValues[3]);

    mapIntToUTF8(0x200000,  &startValues[4]);
    mapIntToUTF8(0x3FFFFFF, &endValues[4]);

    mapIntToUTF8(0x4000000,  &startValues[5]);
    mapIntToUTF8(0x7FFFFFFF, &endValues[5]);

#ifdef DEBUG
    for (i = 0; i < UTF8_MAX_LEN; i++) {
        const utf8 *start = &startValues[i];
        const utf8 *end   = &endValues[i];

        fprintf(stderr, "range %d: [%d: %02X %02X %02X %02X %02X %02X] -  [%d: %02X %02X %02X %02X %02X %02X]\n",
                i+1,
                start->length,
                start->bytes[0], start->bytes[1], start->bytes[2], start->bytes[3], start->bytes[4], start->bytes[5],
                end->length,
                end->bytes[0], end->bytes[1], end->bytes[2], end->bytes[3], end->bytes[4], end->bytes[5]);
    }
#endif

    // Go forth and convert...
    yylex();
}
